# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for helping in crash analysis."""

import re

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.system import environment

ASSERT_CRASH_ADDRESSES = [
    0x0000bbadbeef,
    0x0000fbadbeef,
    0x00001f75b7dd,
    0x0000977537dd,
    0x00009f7537dd,
]
GENERIC_CRASH_TYPES = [
    'Null-dereference', 'Null-dereference READ', 'Null-dereference WRITE',
    'READ', 'UNKNOWN', 'UNKNOWN READ', 'UNKNOWN WRITE', 'WRITE',
    'Uncaught exception'
]
SIGNAL_SIGNATURES_NOT_SECURITY = [
    'Sanitizer: ABRT',
    'Sanitizer: BUS',
    'Sanitizer: FPE',
    'Sanitizer: ILL',
    'Sanitizer: breakpoint',
]
STACKTRACE_TOOL_MARKERS = [
    'AddressSanitizer',
    'ASAN:',
    'CFI: Most likely a control flow integrity violation;',
    'ERROR: libFuzzer',
    'KASAN:',
    'LeakSanitizer',
    'MemorySanitizer',
    'ThreadSanitizer',
    'UndefinedBehaviorSanitizer',
    'UndefinedSanitizer',
]
STACKTRACE_END_MARKERS = [
    'ABORTING',
    'END MEMORY TOOL REPORT',
    'End of process memory map.',
    'END_KASAN_OUTPUT',
    'SUMMARY:',
    'Shadow byte and word',
    '[end of stack trace]',
    '\nExiting',
    'minidump has been written',
]
UBSAN_RUNTIME_ERROR = ' runtime error: '
UBSAN_CRASH_TYPES_NON_SECURITY = [
    'Divide-by-zero',
    'Float-cast-overflow',
    # We do not name this "Signed-integer-overflow", for the sake of syntax
    # being used in LLVM and C++, as there is no "signed int" type.
    'Integer-overflow',
    'Invalid-bool-value',
    'Invalid-builtin-use',
    'Invalid-enum-value',
    'Invalid-null-argument',
    'Invalid-null-return',
    'Misaligned-address',
    'No-return-value',
    'Pointer-overflow',
    'Potential-null-reference',
    'Undefined-shift',
    # Unsigned integer overflow actually is not UB, but there is an additional
    # flag in UBSan that enables this type of check. Unsigned integer overflow
    # issues may cause some bugs, unless it's not an intended overflow, e.g. in
    # math of crpto libraries.
    'Unsigned-integer-overflow',
    'Unreachable code',
]
UBSAN_CRASH_TYPES_SECURITY = [
    'Bad-cast',
    'Index-out-of-bounds',
    'Incorrect-function-pointer-type',
    'Non-positive-vla-bound-value',
    'Object-size',
]
GOLANG_CRASH_TYPES_NON_SECURITY = [
    'Index out of range',
    'Integer divide by zero',
    'Makeslice: len out of range',
    'Slice bounds out of range',
    'Stack overflow',
]

# Default page size of 4KB.
NULL_DEREFERENCE_BOUNDARY = 0x1000


def address_to_integer(address):
  """Attempt to convert an address from a string (hex) to an integer."""
  try:
    return int(address, 16)
  except:
    return 0


def has_marker(stacktrace, marker_list):
  """Return true if the stacktrace has atleast one marker in the marker list."""
  for marker in marker_list:
    if marker in stacktrace:
      return True

  return False


def ignore_stacktrace(crash_stacktrace):
  """Return whether the stacktrace needs to be ignored."""
  # Filter crash based on search exclude pattern specified in job definition.
  search_excludes = environment.get_value('SEARCH_EXCLUDES')
  if search_excludes and re.search(search_excludes, crash_stacktrace):
    return True

  # Match stacktrace against custom defined blacklist regexes in project config.
  stack_blacklist_regexes = (
      local_config.ProjectConfig().get('stacktrace.stack_blacklist_regexes'))
  if not stack_blacklist_regexes:
    return False

  stack_blacklist_regex = re.compile(
      r'(%s)' % '|'.join(stack_blacklist_regexes))
  for line in crash_stacktrace.splitlines():
    if stack_blacklist_regex.match(line):
      return True
  return False


def is_crash(return_code, console_output):
  """Analyze the return code and console output to see if this was a crash."""
  if not return_code:
    return False

  crash_signature = environment.get_value('CRASH_SIGNATURE')
  if crash_signature:
    return re.search(crash_signature, console_output)

  return True


def is_check_failure_crash(stacktrace):
  """Return true if it a CHECK failure crash."""
  # Android-specific exception patterns.
  if environment.is_android():
    if 'Device rebooted' in stacktrace:
      return True
    if 'JNI DETECTED ERROR IN APPLICATION:' in stacktrace:
      return True
    if re.match(r'.*FATAL EXCEPTION.*:', stacktrace, re.DOTALL):
      return True

    # FIXME: Analyze why this is not working with chrome.
    # If the process has died, it is worthwhile to catch this with even a
    # NULL stack.
    # process_died_regex = (r'.*Process %s.*\(pid [0-9]+\) has died' %
    #                       environment.get_value('PKG_NAME'))
    # if re.match(process_died_regex, stacktrace, re.DOTALL):
    #   return True

    # Application CHECK failure known patterns.
  if re.match(r'.*#\s*Fatal error in', stacktrace, re.DOTALL):
    return True
  if 'Check failed:' in stacktrace:
    return True

  # Memory debugging tool CHECK failure.
  if 'Sanitizer CHECK failed:' in stacktrace:
    return True

  return False


def is_memory_tool_crash(stacktrace):
  """Return true if it is a memory debugging tool crash."""
  # Job-specific generic checks.
  crash_signature = environment.get_value('CRASH_SIGNATURE')
  if crash_signature and re.search(crash_signature, stacktrace):
    return True

  # Android specific check.
  # FIXME: Share this regex with stack_analyzer.
  if (environment.is_android() and
      re.match(r'.*signal.*\(SIG.*fault addr ([^ ]*)', stacktrace, re.DOTALL)):
    return True

  # Check if we have a complete stacktrace by location stacktrace end marker.
  # If not, bail out.
  if not has_marker(stacktrace, STACKTRACE_END_MARKERS):
    return False

  # Check if have a UBSan error.
  if has_ubsan_error(stacktrace):
    return True

  # Check if have a stacktrace start marker.
  if has_marker(stacktrace, STACKTRACE_TOOL_MARKERS):
    return True

  return False


def is_null_dereference(int_address):
  """Check to see if this is a null dereference crash address."""
  return int_address < NULL_DEREFERENCE_BOUNDARY


def is_assert_crash_address(int_address):
  """Check to see if this is an ASSERT crash based on the address."""
  return int_address in ASSERT_CRASH_ADDRESSES


def has_signal_for_non_security_bug_type(stacktrace):
  """Checks if any signal which means not security bug presented."""
  if re.search(r'^[ \t]+#0[ \t]+0x[0-9a-f]+[ \t]+in gsignal ', stacktrace,
               re.MULTILINE):
    return True

  for signature in SIGNAL_SIGNATURES_NOT_SECURITY:
    if signature in stacktrace:
      return True

  return False


def is_security_issue(crash_stacktrace, crash_type, crash_address):
  """Based on unsymbolized crash parameters, determine whether it has security
  consequences or not."""
  # Stack traces of any type can be manually labelled as a security issue.
  if re.search('FuzzerSecurityIssue(Critical|High|Medium|Low)',
               crash_stacktrace):
    return True

  # eip == 0.
  if 'pc (nil) ' in crash_stacktrace:
    return True
  if 'pc 0x00000000 ' in crash_stacktrace:
    return True
  if 'pc 0x000000000000 ' in crash_stacktrace:
    return True

  # JNI security crashes.
  if re.match(
      '.*JNI DETECTED ERROR[^\n]+(deleted|invalid|unexpected|unknown|wrong)',
      crash_stacktrace, re.DOTALL):
    return True

  if crash_type == 'CHECK failure':
    # TODO(ochang): Remove this once we pick up newer builds that distinguish
    # DCHECKs from CHECKs.
    checks_have_security_implication = environment.get_value(
        'CHECKS_HAVE_SECURITY_IMPLICATION', False)
    return checks_have_security_implication

  # Release SECURITY_CHECK in Blink shouldn't be marked as a security bug.
  if crash_type == 'Security CHECK failure':
    return False

  # Debug CHECK failure should be marked with security implications.
  if crash_type in ('Security DCHECK failure', 'DCHECK failure'):
    return True

  # Hard crash, explicitly enforced in code.
  if (crash_type == 'Fatal error' or crash_type == 'Unreachable code' or
      crash_type.endswith('Exception') or crash_type.endswith('CHECK failure')):
    return False

  if crash_type == 'Stack-overflow':
    return False

  if crash_type == 'Fatal-signal':
    return False

  if crash_type == 'Missing-library':
    return False

  if crash_type == 'Overwrites-const-input':
    return False

  # LeakSanitizer, finds memory leaks.
  if '-leak' in crash_type:
    return False

  # ThreadSanitizer, finds data races.
  if 'Data race' in crash_type:
    return False

  # ThreadSanitizer, finds lock order issues.
  if 'Lock-order-inversion' in crash_type:
    return False

  # Unexpected conditions reached in the program.
  if crash_type == 'ASSERT_NOT_REACHED':
    return False

  if crash_type in UBSAN_CRASH_TYPES_SECURITY:
    return True

  if crash_type in UBSAN_CRASH_TYPES_NON_SECURITY:
    return False

  if crash_type in GOLANG_CRASH_TYPES_NON_SECURITY:
    return False

  # Floating point exceptions.
  if crash_type == 'Floating-point-exception':
    return False

  # RUNTIME_ASSERT in V8 (not a crash, but is a sign of an error).
  if crash_type == 'RUNTIME_ASSERT':
    return False

  # Correctness failure in V8.
  if crash_type == 'V8 correctness failure':
    return False

  # By default, any assert crash is a security crash.
  # This behavior can be changed by defining
  # |ASSERTS_HAVE_SECURITY_IMPLICATION| in job definition.
  if crash_type == 'ASSERT' or 'ASSERTION FAILED' in crash_stacktrace:
    asserts_have_security_implication = environment.get_value(
        'ASSERTS_HAVE_SECURITY_IMPLICATION', True)
    return asserts_have_security_implication

  # Timeouts/OOMs.
  if crash_type in ('Timeout', 'Out-of-memory'):
    return False

  # Unexpected exit call in fuzz target.
  if crash_type == 'Unexpected-exit':
    return False

  # Kernel Failures are security bugs
  if crash_type.startswith('Kernel failure'):
    return True

  # No crash type, can't process.
  if not crash_type:
    return False

  if has_signal_for_non_security_bug_type(crash_stacktrace):
    return False

  # Anything we don't understand will be marked as security.
  if crash_type not in GENERIC_CRASH_TYPES:
    return True

  # Crash on an unknown address.
  if crash_type in GENERIC_CRASH_TYPES:
    # If the address is not near null, then we it is highly likely
    # to have security consequences.
    int_address = address_to_integer(crash_address)

    # This indicates that there was no assert, but a hard crash.
    # (as the assert would be caught by checks above). So, it
    # does have any security implication.
    if is_assert_crash_address(int_address):
      return False

    if not is_null_dereference(int_address):
      return True

  return False


def has_ubsan_error(stacktrace):
  """Return a bool whether the process output contains UBSan errors that should
  be handled as crashes. Suppressions file alone does not provide granular
  control, e.g. to ignore left shift of negative value which can cause false
  positives in some projects e.g. Chromium."""
  if UBSAN_RUNTIME_ERROR not in stacktrace:
    return False

  # FIXME: Avoid opening this file on every single call.
  ubsan_ignores_file_path = environment.get_suppressions_file(
      'ubsan', suffix='ignores')
  if not ubsan_ignores_file_path:
    # No ignore file exists or is empty, everything is allowed.
    return True

  with open(ubsan_ignores_file_path) as f:
    ubsan_ignore_signatures = f.read().splitlines()

  for line in stacktrace.splitlines():
    ignore_line = False
    for signature in ubsan_ignore_signatures:
      if signature in line:
        ignore_line = True

    if ignore_line:
      continue

    if UBSAN_RUNTIME_ERROR in line:
      return True

  return False



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for running gsutil."""

import os

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell

# Default timeout for a GSUtil sync.
FILES_SYNC_TIMEOUT = 5 * 60 * 60


def _get_gsutil_path():
  """Get path to gsutil executable.

  Returns:
    Path to gsutil executable on the system.
  """
  gsutil_executable = 'gsutil'
  if environment.platform() == 'WINDOWS':
    gsutil_executable += '.cmd'

  gsutil_directory = environment.get_value('GSUTIL_PATH')
  if not gsutil_directory:
    # Try searching the binary in path.
    gsutil_absolute_path = shell.which(gsutil_executable)
    if gsutil_absolute_path:
      return gsutil_absolute_path

    logs.log_error('Cannot locate gsutil in PATH, set GSUTIL_PATH to directory '
                   'containing gsutil binary.')
    return None

  gsutil_absolute_path = os.path.join(gsutil_directory, gsutil_executable)
  return gsutil_absolute_path


def _multiprocessing_args():
  """Get multiprocessing args for gsutil."""
  if utils.cpu_count() == 1:
    # GSUtil's default thread count is 5 as it assumes the common configuration
    # is many CPUs (GSUtil uses num_cpu processes).
    return ['-o', 'GSUtil:parallel_thread_count=16']

  return []


def _filter_path(path, write=False):
  """Filters path if needed. In local development environment, this uses local
  paths from an emulated GCS instead of real GCS. `write` indicates whether if
  `path` is a GCS write destination and that intermediate paths should be
  automatically created."""
  if not path.startswith(storage.GS_PREFIX):
    # Only applicable to GCS paths.
    return path

  local_buckets_path = environment.get_value('LOCAL_GCS_BUCKETS_PATH')
  if not local_buckets_path:
    return path

  if write:
    local_path = storage.FileSystemProvider(
        local_buckets_path).convert_path_for_write(path)
  else:
    local_path = storage.FileSystemProvider(local_buckets_path).convert_path(
        path)

  return local_path


class GSUtilRunner(object):
  """GSUtil runner."""

  def __init__(self, _process_runner=new_process.ProcessRunner):
    default_gsutil_args = ['-m']
    default_gsutil_args.extend(_multiprocessing_args())

    self.gsutil_runner = _process_runner(
        _get_gsutil_path(), default_args=default_gsutil_args)

  def run_gsutil(self, arguments, quiet=False, **kwargs):
    """Run GSUtil."""
    if quiet:
      arguments = ['-q'] + arguments

    env = os.environ.copy()
    if 'PYTHONPATH' in env:
      # GSUtil may be on Python 3, and our PYTHONPATH breaks it because we're on
      # Python 2.
      env.pop('PYTHONPATH')

    return self.gsutil_runner.run_and_wait(arguments, env=env, **kwargs)

  def rsync(self,
            source,
            destination,
            timeout=FILES_SYNC_TIMEOUT,
            delete=True,
            exclusion_pattern=None):
    """Download corpus files from a GCS url.

    Args:
      source: Source to sync from.
      destination: Destination to sync to.
      timeout: Timeout for GSUtil.
      delete: Whether or not to delete files on disk that don't exist locally.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    # Use 'gsutil -m rsync -r' to download files from GCS bucket.
    sync_corpus_command = ['rsync', '-r']
    if delete:
      sync_corpus_command.append('-d')
    if exclusion_pattern:
      sync_corpus_command.extend(['-x', exclusion_pattern])

    sync_corpus_command.extend([
        _filter_path(source, write=True),
        _filter_path(destination, write=True),
    ])

    return self.run_gsutil(sync_corpus_command, timeout=timeout, quiet=True)

  def download_file(self, gcs_url, file_path, timeout=None):
    """Download a file from GCS."""
    command = ['cp', _filter_path(gcs_url), file_path]
    result = self.run_gsutil(command, timeout=timeout)
    if result.return_code:
      logs.log_error('GSUtilRunner.download_file failed:\nCommand: %s\n'
                     'Url: %s\n'
                     'Output %s' % (result.command, gcs_url, result.output))

    return result.return_code == 0

  def upload_file(self,
                  file_path,
                  gcs_url,
                  timeout=None,
                  gzip=False,
                  metadata=None):
    """Upload a single file to a given GCS url."""
    if not file_path or not gcs_url:
      return False

    command = []
    if metadata:
      for key, value in six.iteritems(metadata):
        command.extend(['-h', key + ':' + value])

    command.append('cp')
    if gzip:
      command.append('-Z')

    command.extend([file_path, _filter_path(gcs_url, write=True)])
    result = self.run_gsutil(command, timeout=timeout)

    # Check result of command execution, log output if command failed.
    if result.return_code:
      logs.log_error('GSUtilRunner.upload_file failed:\nCommand: %s\n'
                     'Filename: %s\n'
                     'Output: %s' % (result.command, file_path, result.output))

    return result.return_code == 0

  def upload_files_to_url(self, file_paths, gcs_url, timeout=None):
    """Upload files to the given GCS url.

    Args:
      file_paths: A sequence of file paths to upload.
      gcs_url: GCS URL to upload files to.
      timeout: Timeout for gsutil.

    Returns:
      A bool indicating whether or not the command succeeded.
    """
    if not file_paths or not gcs_url:
      return False

    # Use 'gsutil -m cp -I' to upload given files.
    sync_corpus_command = ['cp', '-I', _filter_path(gcs_url, write=True)]
    filenames_buffer = '\n'.join(file_paths)

    result = self.run_gsutil(
        sync_corpus_command,
        input_data=filenames_buffer.encode('utf-8'),
        timeout=timeout)

    # Check result of command execution, log output if command failed.
    if result.return_code:
      logs.log_error(
          'GSUtilRunner.upload_files_to_url failed:\nCommand: %s\n'
          'Filenames:%s\n'
          'Output: %s' % (result.command, filenames_buffer, result.output))

    return result.return_code == 0



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler that serves the testcase file."""

import urllib.parse

from flask import request

from clusterfuzz._internal.google_cloud_utils import blobs
from handlers import base_handler
from libs import access
from libs import gcs
from libs import handler
from libs import helpers

# Blob's filename can be very long. We only preview the last N characters.
# A too long file name can cause an issue with `curl` and `wget`.
PREVIEW_BLOB_FILENAME_LENTGH = 20


def get_testcase_blob_info(testcase):
  """Get testcase file in the binary form."""
  blob_key = testcase.minimized_keys
  using_minimized_keys = True

  if not blob_key or blob_key == 'NA':
    blob_key = testcase.fuzzed_keys
    using_minimized_keys = False

  if not blob_key:
    raise helpers.EarlyExitException(
        "The testcase (%d) doesn't have fuzzed keys." % testcase.key.id(), 400)

  blob_key = str(urllib.parse.unquote(blob_key))

  blob_info = blobs.get_blob_info(blob_key)
  return blob_info, using_minimized_keys


def get(self):
  """Get testcase file and write it to the handler."""
  testcase_id = request.get('id')
  testcase = access.check_access_and_get_testcase(testcase_id)

  blob_info, _ = get_testcase_blob_info(testcase)

  save_as_filename = 'testcase-%s-%s' % (
      testcase.key.id(), blob_info.filename[-PREVIEW_BLOB_FILENAME_LENTGH:])

  content_disposition = str('attachment; filename=%s' % save_as_filename)
  return self.serve_gcs_object(blob_info.bucket, blob_info.object_path,
                               content_disposition)


class Handler(base_handler.Handler, gcs.SignedGcsHandler):
  """Handler that gets the testcase file."""

  @handler.get(handler.HTML)
  @handler.oauth
  def get(self):
    """Serve the testcase file."""
    return get(self)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Retry decorator. We separate this part out into its
  own file because we want to avoid importing too many modules; now it can be
  used in more places."""

import functools
import inspect
import sys
import time

from clusterfuzz._internal.metrics import logs


def sleep(seconds):
  """Invoke time.sleep. This is to avoid the flakiness of time.sleep. See:
    crbug.com/770375"""
  time.sleep(seconds)


def get_delay(num_try, delay, backoff):
  """Compute backoff delay."""
  return delay * (backoff**(num_try - 1))


def wrap(retries,
         delay,
         function,
         backoff=2,
         exception_type=Exception,
         retry_on_false=False):
  """Retry decorator for a function."""

  assert delay > 0
  assert backoff >= 1
  assert retries >= 0

  def decorator(func):
    """Decorator for the given function."""
    tries = retries + 1
    is_generator = inspect.isgeneratorfunction(func)
    function_with_type = function
    if is_generator:
      function_with_type += ' (generator)'

    def handle_retry(num_try, exception=None):
      """Handle retry."""
      from clusterfuzz._internal.metrics import monitoring_metrics

      if (exception is None or
          isinstance(exception, exception_type)) and num_try < tries:
        logs.log(
            'Retrying on %s failed with %s. Retrying again.' %
            (function_with_type, sys.exc_info()[1]),
            num=num_try,
            total=tries)
        sleep(get_delay(num_try, delay, backoff))
        return True

      monitoring_metrics.TRY_COUNT.increment({
          'function': function,
          'is_succeeded': False
      })
      logs.log_error(
          'Retrying on %s failed with %s. Raise.' % (function_with_type,
                                                     sys.exc_info()[1]),
          total=tries)
      return False

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
      """Regular function wrapper."""
      from clusterfuzz._internal.metrics import monitoring_metrics

      for num_try in range(1, tries + 1):
        try:
          result = func(*args, **kwargs)
          if retry_on_false and not result:
            if not handle_retry(num_try):
              return result

            continue

          monitoring_metrics.TRY_COUNT.increment({
              'function': function,
              'is_succeeded': True
          })
          return result
        except Exception as e:
          if not handle_retry(num_try, exception=e):
            raise

      return None

    @functools.wraps(func)
    def _generator_wrapper(*args, **kwargs):
      """Generator function wrapper."""
      # This argument is not applicable for generator functions.
      assert not retry_on_false

      from clusterfuzz._internal.metrics import monitoring_metrics

      already_yielded_element_count = 0
      for num_try in range(1, tries + 1):
        try:
          for index, result in enumerate(func(*args, **kwargs)):
            if index >= already_yielded_element_count:
              yield result
              already_yielded_element_count += 1

          monitoring_metrics.TRY_COUNT.increment({
              'function': function,
              'is_succeeded': True
          })
          break
        except Exception as e:
          if not handle_retry(num_try, exception=e):
            raise

    if is_generator:
      return _generator_wrapper
    return _wrapper

  return decorator



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the crash_access library."""
# pylint: disable=protected-access

import unittest

import mock

from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from libs import crash_access
from libs import helpers
from libs.query import base


def _has_access(need_privileged_access=False):
  return not need_privileged_access


class AddScopeTest(unittest.TestCase):
  """Test add_scope."""

  def setUp(self):
    Query = base.Query  # pylint: disable=invalid-name

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.external_users._allowed_entities_for_user',
        'libs.crash_access.get_permission_names',
        'libs.access.has_access',
        'libs.access.get_user_job_type',
        'libs.helpers.get_user_email',
        'libs.query.base.Query',
    ])
    self.params = {}
    self.mock.get_user_job_type.return_value = None
    self.mock.get_user_email.return_value = 'test@test.com'
    self.mock._allowed_entities_for_user.return_value = []
    self.mock.get_permission_names.return_value = []

    def create_query():
      q = mock.create_autospec(Query)
      return q

    self.mock.Query.side_effect = create_query
    self.query = base.Query()

  def test_forbidden(self):
    """Test when user is forbidden."""
    self.mock.has_access.return_value = False
    with self.assertRaises(helpers.EarlyExitException):
      crash_access.add_scope(self.query, self.params, 'security_flag',
                             'job_type', 'fuzzer_name')

  def test_default_global_privileged(self):
    """Test the default filter for globally privileged users."""
    self.mock.has_access.return_value = True
    crash_access.add_scope(self.query, self.params, 'security_flag', 'job_type',
                           'fuzzer_name')

    self.assertTrue(self.params['permissions']['everything'])
    self.assertTrue(self.params['permissions']['isPrivileged'])
    self.assertEqual([], self.params['permissions']['jobs'])
    self.assertFalse([], self.params['permissions']['fuzzers'])

    self.query.union.assert_has_calls([])
    self.query.filter.assert_has_calls([])

  def test_default_domain(self):
    """Test the default filter for domain users."""
    self.mock.has_access.side_effect = _has_access
    crash_access.add_scope(self.query, self.params, 'security_flag', 'job_type',
                           'fuzzer_name')

    self.assertTrue(self.params['permissions']['everything'])
    self.assertFalse(self.params['permissions']['isPrivileged'])
    self.assertEqual([], self.params['permissions']['jobs'])
    self.assertFalse([], self.params['permissions']['fuzzers'])

    self.query.filter.assert_has_calls([])
    self.query.union.assert_called_once_with(mock.ANY)

    q = self.query.union.call_args[0][0]
    q.union.assert_has_calls([])
    q.filter.assert_has_calls([mock.call('security_flag', False)])

  def test_domain_with_job_and_fuzzer(self):
    """Test domain user with job and fuzzer."""
    self.mock.has_access.side_effect = _has_access
    self.mock.get_user_job_type.return_value = 'job'
    self.mock._allowed_entities_for_user.side_effect = [['job2'], ['fuzzer']]
    self.mock.get_permission_names.side_effect = [['perm'], ['perm1']]

    crash_access.add_scope(self.query, self.params, 'security_flag', 'job_type',
                           'fuzzer_name')

    self.assertTrue(self.params['permissions']['everything'])
    self.assertFalse(self.params['permissions']['isPrivileged'])
    self.assertListEqual(['perm', 'job'], self.params['permissions']['jobs'])
    self.assertListEqual(['perm1'], self.params['permissions']['fuzzers'])

    self.query.union.assert_has_calls([])
    self.query.union.assert_called_once_with(mock.ANY, mock.ANY, mock.ANY)

    everything_query = self.query.union.call_args[0][0]
    job_query = self.query.union.call_args[0][1]
    fuzzer_query = self.query.union.call_args[0][2]

    everything_query.union.assert_has_calls([])
    job_query.union.assert_has_calls([])
    fuzzer_query.union.assert_has_calls([])

    everything_query.filter.assert_has_calls(
        [mock.call('security_flag', False)])
    job_query.filter_in.assert_has_calls([
        mock.call('job_type', ['job2', 'job']),
    ])
    fuzzer_query.filter_in.assert_has_calls([
        mock.call('fuzzer_name', ['fuzzer']),
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""group_leader tests."""
import datetime
import unittest

from handlers.cron import group_leader
from handlers.cron import grouper


class GroupLeaderTest(unittest.TestCase):
  """Test chooseLeaders."""

  def test_empty(self):
    """Test empty."""
    testcase_map = {}
    group_leader.choose(testcase_map)
    self.assertDictEqual({}, testcase_map)

  def _make_attributes(self, testcase_id, one_time_crasher_flag, issue_id,
                       group_id, time_in_second):
    attributes = grouper.TestcaseAttributes(testcase_id)
    attributes.is_leader = False
    attributes.one_time_crasher_flag = one_time_crasher_flag
    attributes.issue_id = issue_id
    attributes.group_id = group_id
    attributes.timestamp = datetime.datetime.utcfromtimestamp(time_in_second)

    return attributes

  def test_most_recent(self):
    """Test choosing most recent when there's no high-quality testcase."""
    testcase_map = {
        1: self._make_attributes(1, True, None, 11, 10),
        2: self._make_attributes(2, True, None, 11, 1),
        3: self._make_attributes(3, True, None, 22, 10),
        4: self._make_attributes(4, True, None, 22, 1)
    }
    group_leader.choose(testcase_map)
    self.assertTrue(testcase_map[1].is_leader)
    self.assertFalse(testcase_map[2].is_leader)
    self.assertTrue(testcase_map[3].is_leader)
    self.assertFalse(testcase_map[4].is_leader)

  def test_choose_reproducible(self):
    """Test choosing being reproducible over nothing."""
    testcase_map = {
        1: self._make_attributes(1, True, None, 11, 10),
        2: self._make_attributes(2, False, None, 11, 1),
    }
    group_leader.choose(testcase_map)

    self.assertTrue(testcase_map[2].is_leader)
    self.assertFalse(testcase_map[1].is_leader)

  def test_choose_has_issue_over_reproducible(self):
    """Test choosing having issue over being reproducible."""
    testcase_map = {
        1: self._make_attributes(1, False, None, 22, 10),
        2: self._make_attributes(2, True, '9', 22, 1),
    }
    group_leader.choose(testcase_map)

    self.assertTrue(testcase_map[2].is_leader)
    self.assertFalse(testcase_map[1].is_leader)

  def test_choose_has_issue_and_reproducible_over_anything_else(self):
    """Test choosing having issue and being reproducible over anything else
      e.g. having issue but not reproducible."""
    testcase_map = {
        1: self._make_attributes(1, True, '10', 33, 10),
        2: self._make_attributes(2, False, None, 33, 9),
        3: self._make_attributes(3, False, '10', 33, 8),
    }
    group_leader.choose(testcase_map)

    self.assertTrue(testcase_map[3].is_leader)
    self.assertFalse(testcase_map[1].is_leader)
    self.assertFalse(testcase_map[2].is_leader)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""UI related functions."""

import time

from . import adb


def clear_notifications():
  """Clear all pending notifications."""
  adb.run_shell_command(['service', 'call', 'notification', '1'])


def unlock_screen():
  """Unlocks the screen if it is locked."""
  window_dump_output = adb.run_shell_command(['dumpsys', 'window'])
  if 'mShowingLockscreen=true' not in window_dump_output:
    # Screen is not locked, no work to do.
    return

  # Quick power on and off makes this more reliable.
  adb.run_shell_command(['input', 'keyevent', 'KEYCODE_POWER'])
  adb.run_shell_command(['input', 'keyevent', 'KEYCODE_POWER'])

  # This key does the unlock.
  adb.run_shell_command(['input', 'keyevent', 'KEYCODE_MENU'])

  # Artificial delay to let the unlock to complete.
  time.sleep(1)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for project_setup."""
import ast
import copy
import json
import os
import posixpath
import unittest

import flask
from google.cloud import ndb
import googleapiclient
import mock
import six
import webtest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import mock_config
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import project_setup

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'project_setup_data')

EXISTING_BUCKETS = set(['lib1-logs.clusterfuzz-external.appspot.com'])


def _read_data_file(data_file):
  """Helper function to read the contents of a data file."""
  with open(os.path.join(DATA_DIRECTORY, data_file)) as handle:
    return handle.read()


class MockRequest(object):
  """Mock API request."""

  def __init__(self, raise_exception=False, return_value=None):
    self.raise_exception = raise_exception
    self.return_value = return_value

  def execute(self):
    """Mock execute()."""
    if self.raise_exception:
      raise googleapiclient.errors.HttpError(mock.Mock(status=404), b'')

    return self.return_value


def mock_bucket_get(bucket=None):
  """Mock buckets().get()."""
  if bucket in EXISTING_BUCKETS:
    return MockRequest(False, {'name': 'bucket'})

  return MockRequest(True)


def mock_get_iam_policy(bucket=None):
  """Mock buckets().getIamPolicy()."""
  response = {
      'kind': 'storage#policy',
      'resourceId': 'fake',
      'bindings': [],
      'etag': 'fake'
  }

  if bucket in ('lib1-logs.clusterfuzz-external.appspot.com',
                'lib3-logs.clusterfuzz-external.appspot.com'):
    response['bindings'].append({
        'role': 'roles/storage.objectViewer',
        'members': ['user:user@example.com',]
    })

  return MockRequest(return_value=response)


class CopyingMock(mock.MagicMock):
  """MagicMock that copies arguments."""

  def __call__(self, *args, **kwargs):
    args = copy.deepcopy(args)
    kwargs = copy.deepcopy(kwargs)
    return super().__call__(*args, **kwargs)


def mock_set_iam_policy(bucket=None, body=None):  # pylint: disable=unused-argument
  """Mock buckets().setIamPolicy()."""
  bindings = body['bindings']
  if bindings and 'user:primary@example.com' in bindings[0]['members']:
    return MockRequest(raise_exception=True)

  return MockRequest(return_value=copy.deepcopy(body))


def _mock_get_or_create_service_account(project):
  return {
      'email': project + '@serviceaccount.com',
  }


@test_utils.with_cloud_emulators('datastore', 'pubsub')
class OssFuzzProjectSetupTest(unittest.TestCase):
  """Test project_setup for OSS-Fuzz."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/setup', view_func=project_setup.Handler.as_view('/setup'))
    self.app = webtest.TestApp(flaskapp)

    helpers.patch_environ(self)

    data_types.Job(
        name='libfuzzer_asan_old_job',
        environment_string=('MANAGED = True\n'
                            'PROJECT_NAME = old\n')).put()
    data_types.Job(
        name='libfuzzer_msan_old_job',
        environment_string=('MANAGED = True\n'
                            'PROJECT_NAME = old\n')).put()
    data_types.Job(
        name='afl_asan_old_job',
        environment_string=('MANAGED = True\n'
                            'PROJECT_NAME = old\n')).put()
    data_types.Job(
        name='afl_msan_old_job',
        environment_string=('MANAGED = True\n'
                            'PROJECT_NAME = old\n')).put()
    data_types.Job(name='unmanaged_job', environment_string='').put()

    # Will be removed.
    data_types.ExternalUserPermission(
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL,
        entity_name='libfuzzer_asan_lib1',
        email='willberemoved@example.com').put()

    # Existing CC. Makes sure no duplicates are created.
    data_types.ExternalUserPermission(
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL,
        entity_name='libfuzzer_asan_lib1',
        email='primary@example.com').put()

    # Existing project settings. Should not get modified.
    data_types.OssFuzzProject(id='lib1', name='lib1', cpu_weight=1.5).put()

    # Should get deleted.
    data_types.OssFuzzProject(id='old_lib', name='old_lib').put()

    self.libfuzzer = data_types.Fuzzer(name='libFuzzer', jobs=[])
    self.libfuzzer.data_bundle_name = 'global'
    self.libfuzzer.jobs = ['libfuzzer_asan_old_job', 'libfuzzer_msan_old_job']
    self.libfuzzer.put()

    self.afl = data_types.Fuzzer(name='afl', jobs=[])
    self.afl.data_bundle_name = 'global'
    self.afl.jobs = ['afl_asan_old_job', 'afl_msan_old_job']
    self.afl.put()

    self.honggfuzz = data_types.Fuzzer(name='honggfuzz', jobs=[])
    self.honggfuzz.data_bundle_name = 'global'
    self.honggfuzz.put()

    self.gft = data_types.Fuzzer(name='googlefuzztest', jobs=[])
    self.gft.put()

    helpers.patch(self, [
        'clusterfuzz._internal.config.local_config.ProjectConfig',
        ('get_application_id_2',
         'clusterfuzz._internal.base.utils.get_application_id'),
        'clusterfuzz._internal.google_cloud_utils.storage.build',
        'time.sleep',
        'handlers.base_handler.Handler.is_cron',
        'handlers.cron.project_setup.get_oss_fuzz_projects',
        'handlers.cron.service_accounts.get_or_create_service_account',
        'handlers.cron.service_accounts.set_service_account_roles',
    ])

    self.mock.get_or_create_service_account.side_effect = (
        _mock_get_or_create_service_account)

    self.mock.ProjectConfig.return_value = mock_config.MockConfig({
        'segregate_projects':
            True,
        'project_setup': [{
            'source': 'oss-fuzz',
            'build_type': 'RELEASE_BUILD_BUCKET_PATH',
            'add_info_labels': True,
            'add_revision_mappings': True,
            'build_buckets': {
                'afl': 'clusterfuzz-builds-afl',
                'dataflow': 'clusterfuzz-builds-dataflow',
                'honggfuzz': 'clusterfuzz-builds-honggfuzz',
                'libfuzzer': 'clusterfuzz-builds',
                'libfuzzer_i386': 'clusterfuzz-builds-i386',
                'no_engine': 'clusterfuzz-builds-no-engine',
            }
        }]
    })

  def test_execute(self):
    """Tests executing of cron job."""
    mock_storage = mock.MagicMock()
    mock_storage.buckets().insert().execute.return_value = 'timeCreated'
    self.mock.get_application_id_2.return_value = 'clusterfuzz-external'
    self.mock.build.return_value = mock_storage

    pubsub_client = pubsub.PubSubClient()
    app_id = utils.get_application_id()
    unmanaged_topic_name = pubsub.topic_name(app_id, 'jobs-linux')
    old_topic_name = pubsub.topic_name(app_id, 'jobs-shouldbedeleted')
    old_subscription_name = pubsub.subscription_name(app_id,
                                                     'jobs-shouldbedeleted')
    other_topic_name = pubsub.topic_name(app_id, 'other')

    pubsub_client.create_topic(unmanaged_topic_name)
    pubsub_client.create_topic(old_topic_name)
    pubsub_client.create_topic(other_topic_name)
    pubsub_client.create_subscription(old_subscription_name, old_topic_name)

    self.mock.get_oss_fuzz_projects.return_value = [
        ('lib1', {
            'homepage': 'http://example.com',
            'primary_contact': 'primary@example.com',
            'auto_ccs': [
                'User@example.com',
                'user2@googlemail.com',
            ],
        }),
        ('lib2', {
            'homepage': 'http://example2.com',
            'disabled': True,
            'fuzzing_engines': ['libfuzzer',],
        }),
        ('lib3', {
            'homepage':
                'http://example3.com',
            'sanitizers': [
                'address',
                {
                    'memory': {
                        'experimental': True,
                    },
                },
                'undefined',
            ],
            'auto_ccs':
                'User@example.com',
            'disabled':
                False,
            'fuzzing_engines': ['libfuzzer',],
            'view_restrictions':
                'none',
            'architectures': ['i386', 'x86_64'],
        }),
        ('lib4', {
            'homepage': 'http://example4.com',
            'language': 'go',
            'sanitizers': ['address'],
            'auto_ccs': 'User@example.com',
            'fuzzing_engines': ['none'],
            'blackbox': True,
        }),
        ('lib5', {
            'homepage': 'http://example5.com',
            'sanitizers': ['address'],
            'fuzzing_engines': ['libfuzzer',],
            'experimental': True,
            'selective_unpack': True,
            'main_repo': 'https://github.com/google/main-repo',
        }),
        ('lib6', {
            'homepage': 'http://example6.com',
            'sanitizers': ['address', 'dataflow', 'memory', 'undefined'],
            'fuzzing_engines': ['libfuzzer', 'afl', 'dataflow'],
            'auto_ccs': 'User@example.com',
            'vendor_ccs': ['vendor1@example.com', 'vendor2@example.com'],
        }),
    ]

    mock_storage.buckets().get.side_effect = mock_bucket_get
    mock_storage.buckets().getIamPolicy.side_effect = mock_get_iam_policy
    mock_storage.buckets().setIamPolicy = CopyingMock()
    mock_storage.buckets().setIamPolicy.side_effect = mock_set_iam_policy

    self.app.get('/setup')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_lib1').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib1')
    self.assertEqual(job.platform, 'LIB1_LINUX')
    six.assertCountEqual(self, job.templates,
                         ['engine_asan', 'libfuzzer', 'prune'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib1/lib1-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib1\n'
        'SUMMARY_PREFIX = lib1\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib1/lib1-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib1-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib1-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib1-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib1-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib1,Engine-libfuzzer\n')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_lib2').get()
    self.assertIsNone(job)

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_lib3').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib3')
    self.assertEqual(job.platform, 'LIB3_LINUX')
    six.assertCountEqual(self, job.templates,
                         ['engine_asan', 'libfuzzer', 'prune'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib3/lib3-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib3\n'
        'SUMMARY_PREFIX = lib3\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib3/lib3-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib3-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib3-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib3-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib3-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib3,Engine-libfuzzer\n'
        'ISSUE_VIEW_RESTRICTIONS = none\n')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_i386_lib3').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib3')
    self.assertEqual(job.platform, 'LIB3_LINUX')
    six.assertCountEqual(self, job.templates, ['engine_asan', 'libfuzzer'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds-i386/lib3/lib3-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib3\n'
        'SUMMARY_PREFIX = lib3\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds-i386/lib3/lib3-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib3-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib3-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib3-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib3-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib3,Engine-libfuzzer\n'
        'ISSUE_VIEW_RESTRICTIONS = none\n')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_msan_lib3').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib3')
    self.assertEqual(job.platform, 'LIB3_LINUX')
    six.assertCountEqual(self, job.templates, ['engine_msan', 'libfuzzer'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib3/lib3-memory-([0-9]+).zip\n'
        'PROJECT_NAME = lib3\n'
        'SUMMARY_PREFIX = lib3\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib3/lib3-memory-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib3-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib3-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib3-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib3-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib3,Engine-libfuzzer\n'
        'EXPERIMENTAL = True\n'
        'ISSUE_VIEW_RESTRICTIONS = none\n')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_ubsan_lib3').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib3')
    self.assertEqual(job.platform, 'LIB3_LINUX')
    six.assertCountEqual(self, job.templates, ['engine_ubsan', 'libfuzzer'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib3/lib3-undefined-([0-9]+).zip\n'
        'PROJECT_NAME = lib3\n'
        'SUMMARY_PREFIX = lib3\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib3/lib3-undefined-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib3-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib3-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib3-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib3-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib3,Engine-libfuzzer\n'
        'ISSUE_VIEW_RESTRICTIONS = none\n')

    job = data_types.Job.query(data_types.Job.name == 'afl_asan_lib1').get()
    self.assertIsNotNone(job)
    self.assertEqual(job.project, 'lib1')
    self.assertEqual(job.platform, 'LIB1_LINUX')
    six.assertCountEqual(self, job.templates, ['engine_asan', 'afl'])
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds-afl/lib1/lib1-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib1\n'
        'SUMMARY_PREFIX = lib1\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds-afl/lib1/lib1-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib1-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib1-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib1-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib1-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib1,Engine-afl\n'
        'MINIMIZE_JOB_OVERRIDE = libfuzzer_asan_lib1\n')

    # Engine-less job. Manually managed.
    job = data_types.Job.query(data_types.Job.name == 'asan_lib4').get()
    self.assertIsNone(job)

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_lib5').get()
    self.assertEqual(job.project, 'lib5')
    self.assertEqual(job.platform, 'LIB5_LINUX')
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib5/lib5-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib5\n'
        'SUMMARY_PREFIX = lib5\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib5/lib5-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib5-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib5-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib5-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib5-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib5,Engine-libfuzzer\n'
        'EXPERIMENTAL = True\n'
        'UNPACK_ALL_FUZZ_TARGETS_AND_FILES = False\n'
        'MAIN_REPO = https://github.com/google/main-repo\n')

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_lib6').get()
    self.assertEqual(job.project, 'lib6')
    self.assertEqual(job.platform, 'LIB6_LINUX')
    self.assertEqual(
        job.environment_string, 'RELEASE_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds/lib6/lib6-address-([0-9]+).zip\n'
        'PROJECT_NAME = lib6\n'
        'SUMMARY_PREFIX = lib6\n'
        'MANAGED = True\n'
        'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
        'clusterfuzz-builds/lib6/lib6-address-%s.srcmap.json\n'
        'FUZZ_LOGS_BUCKET = lib6-logs.clusterfuzz-external.appspot.com\n'
        'CORPUS_BUCKET = lib6-corpus.clusterfuzz-external.appspot.com\n'
        'QUARANTINE_BUCKET = lib6-quarantine.clusterfuzz-external.appspot.com\n'
        'BACKUP_BUCKET = lib6-backup.clusterfuzz-external.appspot.com\n'
        'AUTOMATIC_LABELS = Proj-lib6,Engine-libfuzzer\n'
        'DATAFLOW_BUILD_BUCKET_PATH = '
        'gs://clusterfuzz-builds-dataflow/lib6/lib6-dataflow-([0-9]+).zip\n')

    self.maxDiff = None  # pylint: disable=invalid-name

    libfuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'libFuzzer').get()
    six.assertCountEqual(self, libfuzzer.jobs, [
        'libfuzzer_asan_lib1',
        'libfuzzer_asan_lib3',
        'libfuzzer_asan_i386_lib3',
        'libfuzzer_asan_lib5',
        'libfuzzer_msan_lib3',
        'libfuzzer_ubsan_lib1',
        'libfuzzer_ubsan_lib3',
        'libfuzzer_asan_lib6',
        'libfuzzer_msan_lib6',
        'libfuzzer_ubsan_lib6',
    ])

    afl = data_types.Fuzzer.query(data_types.Fuzzer.name == 'afl').get()
    six.assertCountEqual(self, afl.jobs, [
        'afl_asan_lib1',
        'afl_asan_lib6',
    ])

    # Test that old unused jobs are deleted.
    self.assertIsNone(
        data_types.Job.query(
            data_types.Job.name == 'libfuzzer_asan_old_job').get())
    self.assertIsNone(
        data_types.Job.query(
            data_types.Job.name == 'libfuzzer_msan_old_job').get())

    # Unmanaged job should still exist.
    self.assertIsNotNone(
        data_types.Job.query(data_types.Job.name == 'unmanaged_job').get())

    # Test that project settings are created.
    lib1_settings = ndb.Key(data_types.OssFuzzProject, 'lib1').get()
    self.assertIsNotNone(lib1_settings)
    self.assertDictEqual({
        'cpu_weight':
            1.5,
        'name':
            'lib1',
        'disk_size_gb':
            None,
        'service_account':
            'lib1@serviceaccount.com',
        'high_end':
            False,
        'ccs': [
            'primary@example.com', 'user@example.com', 'user2@googlemail.com'
        ],
    }, lib1_settings.to_dict())

    lib2_settings = ndb.Key(data_types.OssFuzzProject, 'lib2').get()
    self.assertIsNone(lib2_settings)

    lib3_settings = ndb.Key(data_types.OssFuzzProject, 'lib3').get()
    self.assertIsNotNone(lib3_settings)
    self.assertDictEqual({
        'cpu_weight': 1.0,
        'name': 'lib3',
        'disk_size_gb': None,
        'service_account': 'lib3@serviceaccount.com',
        'high_end': False,
        'ccs': ['user@example.com'],
    }, lib3_settings.to_dict())

    lib4_settings = ndb.Key(data_types.OssFuzzProject, 'lib4').get()
    self.assertIsNotNone(lib4_settings)
    self.assertDictEqual({
        'cpu_weight': 0.2,
        'name': 'lib4',
        'disk_size_gb': None,
        'service_account': 'lib4@serviceaccount.com',
        'high_end': True,
        'ccs': ['user@example.com'],
    }, lib4_settings.to_dict())

    old_lib_settings = ndb.Key(data_types.OssFuzzProject, 'old_lib').get()
    self.assertIsNone(old_lib_settings)

    mock_storage.buckets().get.assert_has_calls([
        mock.call(bucket='lib1-backup.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib1-corpus.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib1-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib1-logs.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib2-backup.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib2-corpus.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib2-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib2-logs.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib3-backup.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib3-corpus.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib3-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(bucket='lib3-logs.clusterfuzz-external.appspot.com'),
    ])

    mock_storage.buckets().insert.assert_has_calls([
        mock.call(
            body={
                'name': 'lib1-backup.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 100
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={'name': 'lib1-corpus.clusterfuzz-external.appspot.com'},
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib1-quarantine.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 90
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib2-backup.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 100
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={'name': 'lib2-corpus.clusterfuzz-external.appspot.com'},
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib2-quarantine.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 90
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib2-logs.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 14
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib3-backup.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 100
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={'name': 'lib3-corpus.clusterfuzz-external.appspot.com'},
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib3-quarantine.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 90
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
        mock.call(
            body={
                'name': 'lib3-logs.clusterfuzz-external.appspot.com',
                'lifecycle': {
                    'rule': [{
                        'action': {
                            'type': 'Delete'
                        },
                        'condition': {
                            'age': 14
                        }
                    }]
                }
            },
            project='clusterfuzz-external'),
        mock.call().execute(),
    ])

    mock_storage.buckets().setIamPolicy.assert_has_calls([
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:primary@example.com']
                }]
            },
            bucket='lib1-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user2@gmail.com']
                }]
            },
            bucket='lib1-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }]
            },
            bucket='lib1-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='lib1-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:primary@example.com']
                }]
            },
            bucket='lib1-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user2@gmail.com']
                }]
            },
            bucket='lib1-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }]
            },
            bucket='lib1-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='lib1-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role':
                        'roles/storage.objectViewer',
                    'members': [
                        'user:primary@example.com', 'user:user@example.com'
                    ]
                }]
            },
            bucket='lib1-logs.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }]
            },
            bucket='lib1-logs.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='lib1-logs.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:primary@example.com']
                }]
            },
            bucket='lib1-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user2@gmail.com']
                }]
            },
            bucket='lib1-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }]
            },
            bucket='lib1-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': [
                        'user:user2@gmail.com', 'user:user@example.com'
                    ]
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='lib1-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='clusterfuzz-external-deployment'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='test-shared-corpus-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket='test-mutator-plugins-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib1@serviceaccount.com']
                }]
            },
            bucket=u'global-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='lib2-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='lib2-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='lib2-logs.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='lib2-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='clusterfuzz-external-deployment'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='test-shared-corpus-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket='test-mutator-plugins-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib2@serviceaccount.com']
                }]
            },
            bucket=u'global-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }]
            },
            bucket='lib3-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='lib3-backup.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }]
            },
            bucket='lib3-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='lib3-corpus.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='lib3-logs.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }]
            },
            bucket='lib3-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['user:user@example.com']
                }, {
                    'role': 'roles/storage.objectAdmin',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='lib3-quarantine.clusterfuzz-external.appspot.com'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='clusterfuzz-external-deployment'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='test-shared-corpus-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket='test-mutator-plugins-bucket'),
        mock.call(
            body={
                'resourceId':
                    'fake',
                'kind':
                    'storage#policy',
                'etag':
                    'fake',
                'bindings': [{
                    'role': 'roles/storage.objectViewer',
                    'members': ['serviceAccount:lib3@serviceaccount.com']
                }]
            },
            bucket=u'global-corpus.clusterfuzz-external.appspot.com')
    ])

    mappings = data_types.FuzzerJob.query()
    tags_fuzzers_and_jobs = [(m.platform, m.fuzzer, m.job) for m in mappings]
    six.assertCountEqual(self, tags_fuzzers_and_jobs, [
        ('LIB1_LINUX', 'afl', 'afl_asan_lib1'),
        ('LIB1_LINUX', 'libFuzzer', 'libfuzzer_asan_lib1'),
        ('LIB3_LINUX', 'libFuzzer', 'libfuzzer_asan_lib3'),
        ('LIB3_LINUX', 'libFuzzer', 'libfuzzer_asan_i386_lib3'),
        ('LIB3_LINUX', 'libFuzzer', 'libfuzzer_msan_lib3'),
        ('LIB1_LINUX', 'libFuzzer', 'libfuzzer_ubsan_lib1'),
        ('LIB3_LINUX', 'libFuzzer', 'libfuzzer_ubsan_lib3'),
        ('LIB5_LINUX', 'libFuzzer', 'libfuzzer_asan_lib5'),
        ('LIB6_LINUX', 'libFuzzer', 'libfuzzer_asan_lib6'),
        ('LIB6_LINUX', 'libFuzzer', 'libfuzzer_msan_lib6'),
        ('LIB6_LINUX', 'libFuzzer', 'libfuzzer_ubsan_lib6'),
        ('LIB6_LINUX', 'afl', 'afl_asan_lib6'),
        ('LIB1_LINUX', 'honggfuzz', 'honggfuzz_asan_lib1'),
    ])

    all_permissions = [
        entity.to_dict()
        for entity in data_types.ExternalUserPermission.query()
    ]

    six.assertCountEqual(self, all_permissions, [{
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_asan_lib1',
        'email': u'primary@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_ubsan_lib1',
        'email': u'primary@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_ubsan_lib1',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_ubsan_lib1',
        'email': u'user2@googlemail.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_asan_lib1',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_asan_lib1',
        'email': u'user2@googlemail.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'afl_asan_lib1',
        'email': u'primary@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'afl_asan_lib1',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'afl_asan_lib1',
        'email': u'user2@googlemail.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_msan_lib3',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_ubsan_lib3',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'libfuzzer_asan_lib3',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'auto_cc': 1,
        'entity_name': u'asan_lib4',
        'email': u'user@example.com'
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'libfuzzer_asan_i386_lib3',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'libfuzzer_msan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'libfuzzer_ubsan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'libfuzzer_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'afl_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor1@example.com',
        'entity_name': u'libfuzzer_msan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor1@example.com',
        'entity_name': u'libfuzzer_ubsan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor1@example.com',
        'entity_name': u'libfuzzer_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor1@example.com',
        'entity_name': u'afl_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor2@example.com',
        'entity_name': u'libfuzzer_msan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor2@example.com',
        'entity_name': u'libfuzzer_ubsan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor2@example.com',
        'entity_name': u'libfuzzer_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'vendor2@example.com',
        'entity_name': u'afl_asan_lib6',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'primary@example.com',
        'entity_name': u'honggfuzz_asan_lib1',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user@example.com',
        'entity_name': u'honggfuzz_asan_lib1',
        'auto_cc': 1
    }, {
        'entity_kind': 1,
        'is_prefix': False,
        'email': u'user2@googlemail.com',
        'entity_name': u'honggfuzz_asan_lib1',
        'auto_cc': 1
    }])

    expected_topics = [
        'projects/clusterfuzz-external/topics/jobs-linux',
        'projects/clusterfuzz-external/topics/other',
        'projects/clusterfuzz-external/topics/jobs-lib1-linux',
        'projects/clusterfuzz-external/topics/jobs-lib3-linux',
        'projects/clusterfuzz-external/topics/jobs-lib4-linux',
        'projects/clusterfuzz-external/topics/jobs-lib5-linux',
        'projects/clusterfuzz-external/topics/jobs-lib6-linux',
    ]
    six.assertCountEqual(self, expected_topics,
                         list(pubsub_client.list_topics('projects/' + app_id)))

    for topic in expected_topics[2:]:
      lib = posixpath.basename(topic).split('-')[1]
      six.assertCountEqual(self, [
          'projects/clusterfuzz-external/subscriptions/'
          'jobs-{}-linux'.format(lib),
      ], pubsub_client.list_topic_subscriptions(topic))

    self.assertIsNotNone(pubsub_client.get_topic(unmanaged_topic_name))
    self.assertIsNotNone(pubsub_client.get_topic(other_topic_name))
    self.assertIsNone(pubsub_client.get_topic(old_topic_name))
    self.assertIsNone(pubsub_client.get_subscription(old_subscription_name))


URL_RESULTS = ast.literal_eval(_read_data_file('url_results.txt'))


def mock_get_url(url):
  """Mock get_url()."""
  if url not in URL_RESULTS:
    return None

  return URL_RESULTS[url]


class MockRequestsGet(object):
  """Mock requests.get."""

  def __init__(self, url, params=None, auth=None):  # pylint: disable=unused-argument
    if url in URL_RESULTS:
      self.text = URL_RESULTS[url]
      self.status_code = 200
    else:
      self.text = None
      self.status_code = 500


@test_utils.with_cloud_emulators('datastore')
class GetLibrariesTest(unittest.TestCase):
  """Test get_oss_fuzz_projects()."""

  def setUp(self):
    data_types.Config(github_credentials='client_id;client_secret').put()

    helpers.patch(self, ['requests.get'])
    self.mock.get.side_effect = MockRequestsGet

  def test_get_oss_fuzz_projects(self):
    """Tests get_oss_fuzz_projects()."""
    libraries = project_setup.get_oss_fuzz_projects()
    self.assertListEqual(
        sorted(libraries), [('boringssl', {
            'homepage': 'https://boringssl.googlesource.com/boringssl/'
        }), ('curl', {
            'homepage': 'https://curl.haxx.se/',
            'dockerfile': {
                'git': 'fake',
                'path': 'path/Dockerfile',
            }
        })])


def _mock_read_data(path):
  """Mock read_data."""
  if 'dbg' in path:
    return json.dumps({
        'projects': [{
            'build_path': 'gs://bucket-dbg/a-b/%ENGINE%/%SANITIZER%/'
                          '%TARGET%/([0-9]+).zip',
            'name': '//a/b',
            'fuzzing_engines': ['libfuzzer', 'honggfuzz'],
            'sanitizers': ['address']
        }]
    })

  return json.dumps({
      'projects': [
          {
              'build_path':
                  'gs://bucket/a-b/%ENGINE%/%SANITIZER%/%TARGET%/([0-9]+).zip',
              'name':
                  '//a/b',
              'fuzzing_engines': ['libfuzzer', 'honggfuzz'],
              'sanitizers': ['address', 'memory']
          },
          {
              'build_path':
                  'gs://bucket/c-d/%ENGINE%/%SANITIZER%/%TARGET%/([0-9]+).zip',
              'name':
                  '//c/d',
              'fuzzing_engines': ['libfuzzer', 'googlefuzztest'],
              'sanitizers': ['address']
          },
      ]
  })


@test_utils.with_cloud_emulators('datastore', 'pubsub')
class GenericProjectSetupTest(unittest.TestCase):
  """Test generic project setup."""

  def setUp(self):
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/setup', view_func=project_setup.Handler.as_view('/setup'))
    self.app = webtest.TestApp(flaskapp)

    helpers.patch_environ(self)

    data_types.Job(name='old_unmanaged').put()
    data_types.Job(
        name='old_managed',
        environment_string='MANAGED = True\nPROJECT_NAME = old').put()

    self.libfuzzer = data_types.Fuzzer(
        name='libFuzzer', jobs=['old_unmanaged', 'old_managed'])
    self.libfuzzer.put()

    self.afl = data_types.Fuzzer(name='afl', jobs=[])
    self.afl.put()

    self.honggfuzz = data_types.Fuzzer(name='honggfuzz', jobs=[])
    self.honggfuzz.put()

    self.gft = data_types.Fuzzer(name='googlefuzztest', jobs=[])
    self.gft.put()

    helpers.patch(self, [
        'clusterfuzz._internal.config.local_config.ProjectConfig',
        ('get_application_id_2',
         'clusterfuzz._internal.base.utils.get_application_id'),
        'clusterfuzz._internal.google_cloud_utils.storage.build',
        'clusterfuzz._internal.google_cloud_utils.storage.read_data',
        'time.sleep',
        'handlers.base_handler.Handler.is_cron',
    ])

    self.mock.read_data.side_effect = _mock_read_data

    self.mock.ProjectConfig.return_value = mock_config.MockConfig({
        'project_setup': [
            {
                'source': 'gs://bucket/projects.json',
                'build_type': 'FUZZ_TARGET_BUILD_BUCKET_PATH',
                'experimental_sanitizers': ['memory'],
                'build_buckets': {
                    'afl': 'clusterfuzz-builds-afl',
                    'dataflow': 'clusterfuzz-builds-dataflow',
                    'honggfuzz': 'clusterfuzz-builds-honggfuzz',
                    'googlefuzztest': 'clusterfuzz-builds-googlefuzztest',
                    'libfuzzer': 'clusterfuzz-builds',
                    'libfuzzer_i386': 'clusterfuzz-builds-i386',
                    'no_engine': 'clusterfuzz-builds-no-engine',
                },
                'additional_vars': {
                    'all': {
                        'STRING_VAR': 'VAL',
                        'BOOL_VAR': True,
                        'INT_VAR': 0,
                    },
                    'libfuzzer': {
                        'address': {
                            'ASAN_VAR': 'VAL',
                        },
                        'memory': {
                            'MSAN_VAR': 'VAL',
                        }
                    }
                }
            },
            {
                'source': 'gs://bucket-dbg/projects.json',
                'job_suffix': '_dbg',
                'external_config': {
                    'reproduction_topic':
                        'projects/proj/topics/reproduction',
                    'updates_subscription':
                        'projects/proj/subscriptions/updates',
                },
                'build_type': 'FUZZ_TARGET_BUILD_BUCKET_PATH',
                'build_buckets': {
                    'afl': 'clusterfuzz-builds-afl-dbg',
                    'dataflow': 'clusterfuzz-builds-dataflow-dbg',
                    'honggfuzz': 'clusterfuzz-builds-honggfuzz-dbg',
                    'googlefuzztest': 'clusterfuzz-builds-googlefuzztest-dbg',
                    'libfuzzer': 'clusterfuzz-builds-dbg',
                    'libfuzzer_i386': 'clusterfuzz-builds-i386-dbg',
                    'no_engine': 'clusterfuzz-builds-no-engine-dbg',
                },
                'additional_vars': {
                    'all': {
                        'STRING_VAR': 'VAL-dbg',
                        'BOOL_VAR': True,
                        'INT_VAR': 0,
                    },
                    'libfuzzer': {
                        'address': {
                            'ASAN_VAR': 'VAL-dbg',
                        },
                        'memory': {
                            'MSAN_VAR': 'VAL-dbg',
                        }
                    }
                }
            },
        ],
    })

    # Should be deleted.
    job = data_types.Job(
        name='libfuzzer_asan_c-d_dbg', environment_string='MANAGED = True')
    job.put()

  def test_execute(self):
    """Tests executing of cron job."""
    self.app.get('/setup')
    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_a-b').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket/a-b/libfuzzer/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //a/b\nSUMMARY_PREFIX = //a/b\nMANAGED = True\n'
        'ASAN_VAR = VAL\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'libfuzzer', 'prune'],
                         job.templates)
    self.assertEqual(None, job.external_reproduction_topic)
    self.assertEqual(None, job.external_updates_subscription)
    self.assertFalse(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_msan_a-b').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket/a-b/libfuzzer/memory/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //a/b\nSUMMARY_PREFIX = //a/b\nMANAGED = True\n'
        'EXPERIMENTAL = True\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'MSAN_VAR = VAL\n'
        'STRING_VAR = VAL\n', job.environment_string)
    six.assertCountEqual(self, ['engine_msan', 'libfuzzer'], job.templates)
    self.assertEqual(None, job.external_reproduction_topic)
    self.assertEqual(None, job.external_updates_subscription)
    self.assertFalse(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_c-d').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket/c-d/libfuzzer/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //c/d\nSUMMARY_PREFIX = //c/d\nMANAGED = True\n'
        'ASAN_VAR = VAL\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'libfuzzer', 'prune'],
                         job.templates)
    self.assertEqual(None, job.external_reproduction_topic)
    self.assertEqual(None, job.external_updates_subscription)
    self.assertFalse(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_c-d_dbg').get()
    self.assertIsNone(job)

    job = data_types.Job.query(
        data_types.Job.name == 'libfuzzer_asan_a-b_dbg').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket-dbg/a-b/libfuzzer/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //a/b\nSUMMARY_PREFIX = //a/b\nMANAGED = True\n'
        'ASAN_VAR = VAL-dbg\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL-dbg\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'libfuzzer', 'prune'],
                         job.templates)
    self.assertEqual('projects/proj/topics/reproduction',
                     job.external_reproduction_topic)
    self.assertEqual('projects/proj/subscriptions/updates',
                     job.external_updates_subscription)
    self.assertTrue(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'honggfuzz_asan_a-b').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket/a-b/honggfuzz/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //a/b\nSUMMARY_PREFIX = //a/b\nMANAGED = True\n'
        'MINIMIZE_JOB_OVERRIDE = libfuzzer_asan_a-b\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'honggfuzz'], job.templates)
    self.assertEqual(None, job.external_reproduction_topic)
    self.assertEqual(None, job.external_updates_subscription)
    self.assertFalse(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'honggfuzz_asan_a-b_dbg').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket-dbg/a-b/honggfuzz/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //a/b\nSUMMARY_PREFIX = //a/b\nMANAGED = True\n'
        'MINIMIZE_JOB_OVERRIDE = libfuzzer_asan_a-b_dbg\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL-dbg\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'honggfuzz'], job.templates)
    self.assertEqual('projects/proj/topics/reproduction',
                     job.external_reproduction_topic)
    self.assertEqual('projects/proj/subscriptions/updates',
                     job.external_updates_subscription)
    self.assertTrue(job.is_external())

    job = data_types.Job.query(
        data_types.Job.name == 'googlefuzztest_asan_c-d').get()
    self.assertEqual(
        'FUZZ_TARGET_BUILD_BUCKET_PATH = '
        'gs://bucket/c-d/googlefuzztest/address/%TARGET%/([0-9]+).zip\n'
        'PROJECT_NAME = //c/d\nSUMMARY_PREFIX = //c/d\nMANAGED = True\n'
        'BOOL_VAR = True\n'
        'INT_VAR = 0\n'
        'STRING_VAR = VAL\n', job.environment_string)
    six.assertCountEqual(self, ['engine_asan', 'googlefuzztest'], job.templates)
    self.assertEqual(None, job.external_reproduction_topic)
    self.assertEqual(None, job.external_updates_subscription)
    self.assertFalse(job.is_external())

    libfuzzer = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'libFuzzer').get()
    six.assertCountEqual(self, [
        'libfuzzer_asan_a-b',
        'libfuzzer_asan_c-d',
        'libfuzzer_msan_a-b',
        'old_unmanaged',
    ], libfuzzer.jobs)

    afl = data_types.Fuzzer.query(data_types.Fuzzer.name == 'afl').get()
    six.assertCountEqual(self, [], afl.jobs)

    honggfuzz = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'honggfuzz').get()
    six.assertCountEqual(self, [
        'honggfuzz_asan_a-b',
    ], honggfuzz.jobs)

    gft = data_types.Fuzzer.query(
        data_types.Fuzzer.name == 'googlefuzztest').get()
    six.assertCountEqual(self, ['googlefuzztest_asan_c-d'], gft.jobs)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Service account creation/role helpers."""
import logging

import googleapiclient

from clusterfuzz._internal.base import utils

_ACCOUNT_PREFIX = 'bot-'
_MIN_LEN = 6
_MAX_LEN = 30
_HASH_PREFIX_LEN = _MAX_LEN - len(_ACCOUNT_PREFIX)


def _create_client(service_name, version='v1'):
  """Create a googleapiclient client."""
  return googleapiclient.discovery.build(service_name, version)


def _service_account_email(project_id, service_account_id):
  """Return full service account email."""
  return '%s@%s.iam.gserviceaccount.com' % (service_account_id, project_id)


def _service_account_id(project):
  """Return service account ID for project."""
  # From
  # cloud.google.com/iam/reference/rest/v1/projects.serviceAccounts/create:
  #
  # The account id that is used to generate the service account email address
  # and a stable unique id. It is unique within a project, must be 6-30
  # characters long, and match the regular expression [a-z]([-a-z0-9]*[a-z0-9])
  # to comply with RFC1035.
  account_id = _ACCOUNT_PREFIX + project.replace('_', '-')
  if not account_id[-1].isalnum():
    # Must end in '[a-z][0-9]'.
    account_id += '0'

  if len(account_id) < _MIN_LEN:
    # Must be at least |min_len| in length.
    account_id = account_id.ljust(_MIN_LEN, '0')

  # Use a hash prefix as the service account name if the project name is too
  # long.
  if len(account_id) > _MAX_LEN:
    account_id = _ACCOUNT_PREFIX + utils.string_hash(project)[:_HASH_PREFIX_LEN]

  assert len(account_id) >= _MIN_LEN and len(account_id) <= _MAX_LEN
  return account_id


def get_service_account(iam, project_id, service_account_id):
  """Try to get a service account. Returns None if it does not exist."""
  try:
    request = iam.projects().serviceAccounts().get(
        name='projects/{0}/serviceAccounts/{1}'.format(
            project_id, _service_account_email(project_id, service_account_id)))

    return request.execute()
  except googleapiclient.errors.HttpError as e:
    if e.resp.status == 404:
      return None

    raise


def get_or_create_service_account(project):
  """Get or create service account for the project."""
  iam = _create_client('iam')
  project_id = utils.get_application_id()
  service_account_id = _service_account_id(project)

  service_account = get_service_account(iam, project_id, service_account_id)
  if service_account:
    logging.info('Using existing new service account for %s.', project)
    return service_account

  logging.info('Creating new service account for %s.', project)
  request = iam.projects().serviceAccounts().create(
      name='projects/' + project_id,
      body={
          'accountId': service_account_id,
          'serviceAccount': {
              'displayName': project,
          }
      })

  return request.execute()


def _get_or_insert_iam_binding(policy, role):
  """Return the binding corresponding to the given role. Creates the binding if
  needed."""
  existing_binding = next(
      (binding for binding in policy['bindings'] if binding['role'] == role),
      None)
  if existing_binding:
    return existing_binding

  new_binding = {
      'role': role,
      'members': [],
  }

  policy['bindings'].append(new_binding)
  return new_binding


def _add_service_account_role(policy, role, service_account):
  """Add a role to a service account. Returns whether or not changes were
  made."""
  binding = _get_or_insert_iam_binding(policy, role)

  service_account_member = 'serviceAccount:' + service_account
  if service_account_member not in binding['members']:
    binding['members'].append(service_account_member)
    return True

  return False


def set_service_account_roles(service_account):
  """Set roles for service account."""
  project_id = utils.get_application_id()
  resource_manager = _create_client('cloudresourcemanager')

  request = resource_manager.projects().getIamPolicy(
      resource=project_id, body={})
  policy = request.execute()

  # Set logging and metrics permissions.
  policy_changed = False
  policy_changed |= _add_service_account_role(policy, 'roles/logging.logWriter',
                                              service_account['email'])
  policy_changed |= _add_service_account_role(
      policy, 'roles/monitoring.metricWriter', service_account['email'])

  if not policy_changed:
    return

  request = resource_manager.projects().setIamPolicy(
      resource=project_id, body={
          'policy': policy,
      })
  request.execute()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the issue redirector handler."""

import unittest

import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class HandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.issue_management.issue_tracker_utils.get_issue_url',
        'libs.helpers.get_testcase',
        'clusterfuzz._internal.metrics.logs._is_running_on_app_engine',
    ])
    self.mock._is_running_on_app_engine.return_value = True  # pylint: disable=protected-access

    import server
    self.app = webtest.TestApp(server.app)

  def test_succeed(self):
    """Test redirection succeeds."""
    testcase = data_types.Testcase()
    testcase.bug_information = '456789'
    self.mock.get_testcase.return_value = testcase
    self.mock.get_issue_url.return_value = 'http://google.com/456789'

    response = self.app.get('/issue/12345')

    self.assertEqual(302, response.status_int)
    self.assertEqual('http://google.com/456789', response.headers['Location'])

    self.mock.get_testcase.assert_has_calls([mock.call('12345')])
    self.mock.get_issue_url.assert_has_calls([mock.call(testcase)])

  def test_no_issue_url(self):
    """Test no issue url."""
    self.mock.get_testcase.return_value = data_types.Testcase()
    self.mock.get_issue_url.return_value = ''

    response = self.app.get('/issue/12345', expect_errors=True)
    self.assertEqual(404, response.status_int)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flash related functions."""
import datetime
import os
import socket
import time

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.datastore import locks
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

from . import adb
from . import constants
from . import fetch_artifact
from . import settings

FLASH_IMAGE_REGEXES = [
    r'.*[.]img',
    r'.*-img-.*[.]zip',
]
FLASH_CUTTLEFISH_REGEXES = [
    r'.*-img-.*[.]zip',
    r'cvd-host_package.tar.gz',
]
FLASH_IMAGE_FILES = [
    # Order is important here.
    ('bootloader', 'bootloader*.img'),
    ('radio', 'radio*.img'),
    ('boot', 'boot.img'),
    ('system', 'system.img'),
    ('recovery', 'recovery.img'),
    ('vendor', 'vendor.img'),
    ('cache', 'cache.img'),
    ('vbmeta', 'vbmeta.img'),
    ('dtbo', 'dtbo.img'),
    ('userdata', 'userdata.img'),
]
FLASH_INTERVAL = 1 * 24 * 60 * 60
FLASH_RETRIES = 3
FLASH_REBOOT_BOOTLOADER_WAIT = 15
FLASH_REBOOT_WAIT = 5 * 60


def download_latest_build(build_info, image_regexes, image_directory):
  """Download the latest build artifact for the given branch and target."""
  # Check if our local build matches the latest build. If not, we will
  # download it.
  build_id = build_info['bid']
  target = build_info['target']
  last_build_info = persistent_cache.get_value(constants.LAST_FLASH_BUILD_KEY)
  if last_build_info and last_build_info['bid'] == build_id:
    return

  # Clean up the images directory first.
  shell.remove_directory(image_directory, recreate=True)
  for image_regex in image_regexes:
    image_file_path = fetch_artifact.get(build_id, target, image_regex,
                                         image_directory)
    if not image_file_path:
      logs.log_error('Failed to download artifact %s for '
                     'branch %s and target %s.' %
                     (image_file_path, build_info['branch'], target))
      return
    if image_file_path.endswith('.zip') or image_file_path.endswith('.tar.gz'):
      archive.unpack(image_file_path, image_directory)


def flash_to_latest_build_if_needed():
  """Wipes user data, resetting the device to original factory state."""
  if environment.get_value('LOCAL_DEVELOPMENT'):
    # Don't reimage local development devices.
    return

  run_timeout = environment.get_value('RUN_TIMEOUT')
  if run_timeout:
    # If we have a run timeout, then we are already scheduled to bail out and
    # will be probably get re-imaged. E.g. using frameworks like Tradefed.
    return

  # Check if a flash is needed based on last recorded flash time.
  last_flash_time = persistent_cache.get_value(
      constants.LAST_FLASH_TIME_KEY,
      constructor=datetime.datetime.utcfromtimestamp)
  needs_flash = last_flash_time is None or dates.time_has_expired(
      last_flash_time, seconds=FLASH_INTERVAL)
  if not needs_flash:
    return

  is_google_device = settings.is_google_device()
  if is_google_device is None:
    logs.log_error('Unable to query device. Reimaging failed.')
    adb.bad_state_reached()

  elif not is_google_device:
    # We can't reimage these, skip.
    logs.log('Non-Google device found, skipping reimage.')
    return

  # Check if both |BUILD_BRANCH| and |BUILD_TARGET| environment variables
  # are set. If not, we don't have enough data for reimaging and hence
  # we bail out.
  branch = environment.get_value('BUILD_BRANCH')
  target = environment.get_value('BUILD_TARGET')
  if not target:
    # We default to userdebug configuration.
    build_params = settings.get_build_parameters()
    if build_params:
      target = build_params.get('target') + '-userdebug'

      # Cache target in environment. This is also useful for cases when
      # device is bricked and we don't have this information available.
      environment.set_value('BUILD_TARGET', target)

  if not branch or not target:
    logs.log_warn(
        'BUILD_BRANCH and BUILD_TARGET are not set, skipping reimage.')
    return

  image_directory = environment.get_value('IMAGES_DIR')
  build_info = fetch_artifact.get_latest_artifact_info(branch, target)
  if not build_info:
    logs.log_error('Unable to fetch information on latest build artifact for '
                   'branch %s and target %s.' % (branch, target))
    return

  if environment.is_android_cuttlefish():
    download_latest_build(build_info, FLASH_CUTTLEFISH_REGEXES, image_directory)
    adb.recreate_cuttlefish_device()
    adb.connect_to_cuttlefish_device()
  else:
    download_latest_build(build_info, FLASH_IMAGE_REGEXES, image_directory)
    # We do one device flash at a time on one host, otherwise we run into
    # failures and device being stuck in a bad state.
    flash_lock_key_name = 'flash:%s' % socket.gethostname()
    if not locks.acquire_lock(flash_lock_key_name, by_zone=True):
      logs.log_error('Failed to acquire lock for reimaging, exiting.')
      return

    logs.log('Reimaging started.')
    logs.log('Rebooting into bootloader mode.')
    for _ in range(FLASH_RETRIES):
      adb.run_as_root()
      adb.run_command(['reboot-bootloader'])
      time.sleep(FLASH_REBOOT_BOOTLOADER_WAIT)
      adb.run_fastboot_command(['oem', 'off-mode-charge', '0'])
      adb.run_fastboot_command(['-w', 'reboot-bootloader'])

      for partition, partition_image_filename in FLASH_IMAGE_FILES:
        partition_image_file_path = os.path.join(image_directory,
                                                 partition_image_filename)
        adb.run_fastboot_command(
            ['flash', partition, partition_image_file_path])
        if partition in ['bootloader', 'radio']:
          adb.run_fastboot_command(['reboot-bootloader'])

      # Disable ramdump to avoid capturing ramdumps during kernel crashes.
      # This causes device lockup of several minutes during boot and we intend
      # to analyze them ourselves.
      adb.run_fastboot_command(['oem', 'ramdump', 'disable'])

      adb.run_fastboot_command('reboot')
      time.sleep(FLASH_REBOOT_WAIT)

      if adb.get_device_state() == 'device':
        break
      logs.log_error('Reimaging failed, retrying.')

    locks.release_lock(flash_lock_key_name, by_zone=True)

  if adb.get_device_state() != 'device':
    logs.log_error('Unable to find device. Reimaging failed.')
    adb.bad_state_reached()

  logs.log('Reimaging finished.')

  # Reset all of our persistent keys after wipe.
  persistent_cache.delete_value(constants.BUILD_PROP_MD5_KEY)
  persistent_cache.delete_value(constants.LAST_TEST_ACCOUNT_CHECK_KEY)
  persistent_cache.set_value(constants.LAST_FLASH_BUILD_KEY, build_info)
  persistent_cache.set_value(constants.LAST_FLASH_TIME_KEY, time.time())
